

#include "generalization.h"
#include "mech.h"
#include "mechDistortion.h"
#include "calCred.h"



// 1: true
// 0: false
int isAlphaKAnonymity(data *dataSet, summary *synopsis, configure *config)
{
	int i;

	int noOfSen;

	int totalNo;

	//noOfSen = 0;

	// sorting according to the list of attributes in config->senAttrNo
	//data_sort(dataSet, config->noOfSenAttr, config->senAttrNo);
	//summary_sort(synopsis, config->noOfSenAttr, config->senAttrNo);

	// Assume only one sensitve attribute
	//int summary_getNoOfSen(data *dataSet, summary *synopsis, int noOfDim, int *dimNo)
	noOfSen = summary_getNoOfSen(dataSet, synopsis, config->noOfSenAttr, config->senAttrNo);

	// can be replaced with totalNo = dataSet->noOfTuple
	totalNo = 0;
	for (i = 0; i < synopsis->noOfGroup; i++)
	{
		totalNo += synopsis->grouping[i].noOfMember;
	}

	if (noOfSen/(double)totalNo <= config->alpha)
	{
		return 1;
	}
	else
	{
		return 0;
	}
	

}


void generalizeDimension(dimension *dim, generalizedLevelTable *genLevelTable)
{
	int i;

	int noOfPoss;


	noOfPoss = genLevelTable->noOfPoss;

	for (i = 0; i < noOfPoss; i++)
	{
		if (genLevelTable->gen[i].bottomValue == dim->value)
		{
			// generalize
			dim->value = genLevelTable->gen[i].topValue;
			dim->level = genLevelTable->gen[i].topLevel;

			break;
		}
	}
}

void generalizeEquivalenceTuple(equivalenceTuple *eqTuple, int attrNo, generalizedLevelTable *genLevelTable)
{
	dimension *dim;
	
	dim = &(eqTuple->content[attrNo]);
	generalizeDimension(dim, genLevelTable);
}

void generalizeEqClass(equivalenceClass *eqClass, int attrNo, configure *conf, generalizedAllTable *genAllTable)
{
	int i;

	int noOfDim;
	int genBottomLevel;

	equivalenceTuple **eqTuple;
	equivalenceTuple *eqOne;


	generalizedDimTable *genDimTable;
	generalizedLevelTable *genLevelTable;

	noOfDim = conf->noOfDim;

	genDimTable = &(genAllTable->genDimTable[attrNo]);

	eqTuple = (equivalenceTuple **) eqClass->eqClass; 

	if (eqClass->noOfEqClass > 0)
	{
		genBottomLevel = eqTuple[0]->content[attrNo].level;
		genLevelTable = &(genDimTable->gen[genBottomLevel]);

		for (i = 0; i < eqClass->noOfEqClass; i++)
		{
			//eqOne = &(eqTuple[i]);
			eqOne = eqTuple[i];
			generalizeEquivalenceTuple(eqOne, attrNo, genLevelTable);
		}
	}
	else
	{
		// do nothing
	}
}

void calculateSensitiveCount(equivalenceClass *eqClass, configure *conf, int *senCount, int *totalCount)
{
	int i;

	int inSenCount;
	int inTotalCount;

	equivalenceTuple **eqTuple;
	equivalenceTuple *eqOne;

	eqTuple = eqClass->eqClass;

	inSenCount = 0;
	inTotalCount = 0;
	for (i = 0; i < eqClass->noOfEqClass; i++)
	{
		//eqOne = &(eqTuple[i]);
		eqOne = eqTuple[i];

		inTotalCount += eqOne->noOfMember;
		inSenCount += eqOne->noOfSen;
	}

	(*senCount) = inSenCount;
	(*totalCount) = inTotalCount;
}

int calculateGreatestPossRemovalCount(equivalenceClass *eqClass, configure *conf)
{
	int returnValue;

	int senCount;
	int totalCount;

	calculateSensitiveCount(eqClass, conf, &senCount, &totalCount);

	// flooring
	returnValue = (int) (totalCount - senCount/conf->alpha);

	if (returnValue < 0)
	{
		returnValue = 0;
	}

	return returnValue;
}


void fullSpaceGeneralize(equivalenceClass *projEqClass, data *dataSet, configure *conf, generalizedAllTable *genAllTable, AttributeList *attrList, GeneralizationDomain *genDomain)
{
	int i, j;
	int noOfDim;

	int *list;
	int noOfAttr;

	int *genStrength;
	int *genLevel;

	int dimNo;

	noOfDim = conf->noOfDim-1;

	genLevel = genDomain->genLevel;

	// preprocess
	noOfAttr = attrList->noOfAttr;
	list = attrList->attr;

	genStrength = (int *) malloc(sizeof(int)*noOfDim);
	memset(genStrength, 0, sizeof(int)*noOfDim);
	for (i = 0; i < noOfAttr; i++)
	{
		dimNo = list[i];

		genStrength[dimNo] = genLevel[i];
	}
	

	// generalize
	for (i = 0; i < noOfDim; i++)
	{
		dimNo = i;
		for (j = 0; j < genStrength[dimNo]; j++)
		{
			generalizeEqClass(projEqClass, dimNo, conf, genAllTable);
		}
	}

	// merge
	mergeEqClass(projEqClass, dataSet, conf);

	free(genStrength);
}


int isSatisfyKAnonymity(equivalenceClass *eqClass, configure *conf)
{
	int noOfEqTuple;
	equivalenceTuple **eqTuple;

	int returnValue;

	equivalenceTuple *oneEqTuple;

	int noOfMember, noOfSen;

	double alpha;

	int K;

	int i;

	K = conf->K;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTuple = eqClass->eqClass;

	alpha = conf->alpha;
	
	returnValue = 1;
	for (i = 0; i < noOfEqTuple; i++)
	{
		oneEqTuple = eqTuple[i];

		noOfMember = oneEqTuple->noOfMember;
		noOfSen = oneEqTuple->noOfSen;

		// *** IMPORTATN change for m-deassociation
		//if ((noOfMember < K) || (noOfSen > ceiling(alpha*noOfMember)))
		if ((noOfMember < K) || (noOfSen > (int) (alpha*noOfMember)))
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

double *checkAllGenDomainSet(AllGeneralizationDomainSet *allGenDomainSet, equivalenceClass *eqClass, configure *conf, data *dataSet, generalizedAllTable *genAllTable, int isCalculate)
{
	double *returnValue;

	int noOfSet;
	GeneralizationDomainSet *set;

	int canReturn;
	
	equivalenceClass *projEqClass;
	equivalenceClass *cloneProjEqClass;

	AttributeList *attrList;

	int i, j;

	int noOfDim, noOfDomain;

	int isSatisfy;

	double distortion;

	GeneralizationDomain *domain;
	GeneralizationDomain *genDomain;
	GeneralizationDomainSet *genDomainSet;


	noOfSet = allGenDomainSet->noOfSet;
	set = allGenDomainSet->set;

	noOfDim = conf->noOfDim;

	if (isCalculate == 1)
	{
		if (noOfSet == 1)
		{
			canReturn = 1;
		}
		else
		{
			canReturn = 0;
		}
	}
	else
	{
		canReturn = 0;
	}


	for (i = 0; i < noOfSet; i++)
	{ 
		genDomainSet = &(set[i]);


		// create the projected eq. classes
		attrList = &(genDomainSet->attrList);
		projEqClass = cloneEquivalenceClassWithProjection(eqClass, noOfDim, attrList, dataSet, conf);

		//fprintf(g_debugFilePtr, "############## eqClass ##################\n");
		//printf("############## eqClass ##################\n");
		//printEquivalenceClass(g_debugFilePtr, eqClass, conf->noOfDim);

		//fprintf(g_debugFilePtr, "############## projEqClass ##################\n");
		//printf("############## projEqClass ##################\n");
		//printEquivalenceClass(g_debugFilePtr, projEqClass, conf->noOfDim);

		domain = genDomainSet->domain;
		noOfDomain = genDomainSet->noOfDomain;

		// returnValue
		if (canReturn == 1)
		{
			returnValue = (double *) malloc(sizeof(double)*noOfDomain);
			for (j = 0; j < noOfDomain ; j++)
			{
				returnValue[j] = -1;
			}
		}

		for (j = 0; j < noOfDomain; j++)
		{
			genDomain = &(domain[j]);

			if (genDomain->isAlpha == 0)
			{
				// clone
				cloneProjEqClass = cloneEquivalenceClass(projEqClass, conf);

				//fprintf(g_debugFilePtr, "############## cloneProjEqClass (before) ##################\n");
				//printf("############## cloneProjEqClass (before) ##################\n");
				//printEquivalenceClass(g_debugFilePtr, cloneProjEqClass, conf->noOfDim);

				// generalize
				fullSpaceGeneralize(cloneProjEqClass, dataSet, conf, genAllTable, attrList, genDomain);

				//fprintf(g_debugFilePtr, "############## cloneProjEqClass (after) ##################\n");
				//printf("############## cloneProjEqClass (after) ##################\n");
				//printEquivalenceClass(g_debugFilePtr, cloneProjEqClass, conf->noOfDim);

				// check isAlpha
				isSatisfy = isSatisfyKAnonymity(cloneProjEqClass, conf);
				if (isSatisfy == 1)
				{
					//
					genDomain->isAlpha = 1;
					markAllParent(genDomainSet, j);

					if (isCalculate == 1)
					{
						// calculate the distortion
						distortion = calculateDistortion(cloneProjEqClass, conf, genAllTable);

						returnValue[j] = distortion;
					}
				}

				// free
				freeEquivalenceClass(cloneProjEqClass);
				free(cloneProjEqClass);
			}
			else
			{
				// do nothing
			}
		}

		freeEquivalenceClass(projEqClass);
		free(projEqClass);
	}

	if (canReturn == 0)
	{
		returnValue = NULL;
	}

	return returnValue;

}


// no. of domain <= 0
int chooseBestGenDomain(GeneralizationDomainSet *genDomainSet, double *dist)
{
	int returnValue;
	int noOfDomain;
	GeneralizationDomain *domain;

	double minDist; 

	int i;

	returnValue = -1;

	domain = genDomainSet->domain;
	noOfDomain = genDomainSet->noOfDomain;

	if (noOfDomain > 0)
	{
		//minDist = dist[0];
		//returnValue = -1;
		minDist = -1;
		for (i = 0; i < noOfDomain; i++)
		{
			if ((dist[i] != -1) && ((dist[i] < minDist) || (minDist == -1)))
			{
				minDist = dist[i];
				returnValue = i;
			}
		}
	}
	else
	{
		returnValue = -1;
	}

	return returnValue;
	
}

equivalenceClass *chooseBestFinalEqClass(AllGeneralizationDomainSet *allGenDomainSet, double *dist, equivalenceClass *eqClass, data *dataSet, configure *conf, generalizedAllTable *genAllTable)
{
	equivalenceClass *returnValue;

	GeneralizationDomainSet *genDomainSet;

	GeneralizationDomain *genDomain;
	AttributeList *attrList;

	int noOfSet;
	GeneralizationDomainSet *set;

	int bestIndex;

	noOfSet = allGenDomainSet->noOfSet;
	set = allGenDomainSet->set;

	if (noOfSet == 1)
	{
		genDomainSet = &(set[0]);

		// find best
		bestIndex = chooseBestGenDomain(genDomainSet, dist);

		if (bestIndex == -1)
		{
			returnValue = NULL;
		}
		else
		{
			// clone
			returnValue = cloneEquivalenceClass(eqClass, conf);

			// generalize
			attrList = &(genDomainSet->attrList);
			genDomain = &(genDomainSet->domain[bestIndex]);
			fullSpaceGeneralize(returnValue, dataSet, conf, genAllTable, attrList, genDomain);
		}
	}
	else
	{
		returnValue = NULL;
	}

	
	return returnValue;
}

int main()
{
	char *configFilename = "conf.txt";

	configure *config;
	data *dataSet;
	generalizedAllTable *genAllTable;

	config = readConfig(configFilename);
	dataSet = readData(config);
	genAllTable = readGen(config);

	printf("Calculating the coverage...\n");
	//void calculateCoverage(generalizedAllTable *genAllTable, int noOfDim)
	calculateCoverage(genAllTable, config->noOfDim);

	//void progressiveMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable)
	//progressiveMain(dataSet, config, genAllTable);

	//hierachicalMain(dataSet, config, genAllTable);
	//randomMain(dataSet, config, genAllTable);
	//fullSpaceMain(dataSet, config, genAllTable);

	//topDownAlphaMain(dataSet, config, genAllTable);

	//mechMain(dataSet, config, genAllTable);
	//mechDistortionMain(dataSet, config, genAllTable);

	calCredMain(dataSet, config, genAllTable);

	//testQsort();
}
