

#include "mech.h"


//int mechMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable)
//{
//	return 0;
//}
 
mechGroup *mechGroup_new(int noOfSenValue)
{
	mechGroup *returnValue;

	returnValue = (mechGroup *) malloc(sizeof(mechGroup));
	memset(returnValue, 0, sizeof(mechGroup));

	returnValue->noOfEachSen = (int *) malloc(sizeof(int)*noOfSenValue);
	memset(returnValue->noOfEachSen, 0, sizeof(int)*noOfSenValue);

	return returnValue;
}

void mechGroup_free(mechGroup *aMechGroup)
{
	int noOfMember;

	noOfMember = aMechGroup->noOfMember;
	if (noOfMember > 0)
	{
		free(aMechGroup->isSen);
		free(aMechGroup->member);
	}

	free(aMechGroup->noOfEachSen);
	free(aMechGroup);
}

void mechGroup_print(FILE *filePtr, mechGroup *aMechGroup)
{
	int i;

	fprintf(filePtr, "    No Of Member: %d\n", aMechGroup->noOfMember);
	fprintf(filePtr, "    No Of Sen   : %d\n", aMechGroup->noOfSen);

	fprintf(filePtr, "    member: ");
	for (i = 0; i < aMechGroup->noOfMember; i++)
	{
		fprintf(filePtr, "%d ", aMechGroup->member[i]);
	}
	fprintf(filePtr, "\n");

	fprintf(filePtr, "    isSen : ");
	for (i = 0; i < aMechGroup->noOfMember; i++)
	{
		fprintf(filePtr, "%d ", aMechGroup->isSen[i]);
	}
	fprintf(filePtr, "\n");

}

void mech_insert(mechGroup *aMechGroup, int id, int isSen, int senValue)
{
	int noOfMember;

	noOfMember = aMechGroup->noOfMember;

	if (noOfMember == 0)
	{
		aMechGroup->member = (int *) malloc(sizeof(int));
		aMechGroup->isSen = (int *) malloc(sizeof(int));
	}
	else
	{
		aMechGroup->member = (int *) realloc(aMechGroup->member, sizeof(int)*(noOfMember+1));
		aMechGroup->isSen = (int *) realloc(aMechGroup->isSen, sizeof(int)*(noOfMember+1));

	}

	aMechGroup->member[noOfMember] = id;
	aMechGroup->isSen[noOfMember] = isSen;
	aMechGroup->noOfMember++;
	if (isSen == 1)
	{
		aMechGroup->noOfSen++;
	}
	aMechGroup->noOfEachSen[senValue]++;
}


int isQIDEqual(int id1, int id2, data *dataSet, int noOfQID)
{
	int returnValue;

	int j;

	tuple *tupleList;

	tuple *tuple1;
	tuple *tuple2;

	tupleList = dataSet->tupleList;

	tuple1 = &(tupleList[id1]);
	tuple2 = &(tupleList[id2]);

	returnValue = 1;
	for (j = 0; j < noOfQID; j++)
	{
		if (tuple1->content[j].value != tuple2->content[j].value)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

int mech_isAllSameQIDInAMechGroup(mechGroup *aMechGroup, data *dataSet, int noOfQID)
{
	int returnValue;
	int noOfMember; 

	int id1, id2;

	int *member;

	int i, j;

	int isEqual;

	noOfMember = aMechGroup->noOfMember;
	member = aMechGroup->member;

	if (noOfMember == 0)
	{
		returnValue = 0;
		return returnValue;
	}
	else
	{
		returnValue = 1;
	}

	for (i = 0; i < noOfMember; i++)
	{
		for (j = i+1; j < noOfMember; j++)
		{
			id1 = member[i];
			id2 = member[j];

			// compare data i and data j

			// if not equal
			isEqual = isQIDEqual(id1, id2, dataSet, noOfQID);
			if (isEqual == 0)
			{
				returnValue = 0;
				return returnValue;
			}
		}
	}

	return returnValue;
}

// assume aMechGroup1 and aMechGroup2 have the same QID within a group
int mech_isAllSameQIDTwoMechGroup(mechGroup *aMechGroup1, mechGroup *aMechGroup2, data *dataSet, int noOfQID)
{
	int returnValue;

	int id1, id2;

	if ((aMechGroup1->noOfMember > 0) && (aMechGroup2->noOfMember > 0))
	{
		id1 = aMechGroup1->member[0];
		id2 = aMechGroup2->member[0];

		returnValue = isQIDEqual(id1, id2, dataSet, noOfQID);
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}

mechGroup *mech_mergeMechGroup(mechGroup *aMechGroup1, mechGroup *aMechGroup2, int noOfValueInSenAttr)
{
	mechGroup *returnValue;


	int noOfMember1, noOfMember2;

	returnValue = mechGroup_new(noOfValueInSenAttr);

	noOfMember1 = aMechGroup1->noOfMember;
	noOfMember2 = aMechGroup2->noOfMember;

	returnValue->noOfMember = noOfMember1 + noOfMember2;
	returnValue->noOfSen = aMechGroup1->noOfSen + aMechGroup2->noOfSen;

	returnValue->member = (int *) malloc(sizeof(int)*returnValue->noOfMember);
	returnValue->isSen = (int *) malloc(sizeof(int)*returnValue->noOfMember);

	memcpy(returnValue->member, aMechGroup1->member, sizeof(int)*noOfMember1);
	memcpy(&(returnValue->member[noOfMember1]), aMechGroup2->member, sizeof(int)*noOfMember2);

	memcpy(returnValue->isSen, aMechGroup1->isSen, sizeof(int)*noOfMember1);
	memcpy(&(returnValue->isSen[noOfMember1]), aMechGroup2->isSen, sizeof(int)*noOfMember2);

	return returnValue;
}




mechAllGroup *mechAllGroup_new()
{
	mechAllGroup *returnValue;

	returnValue = (mechAllGroup *) malloc(sizeof(mechAllGroup));
	memset(returnValue, 0, sizeof(mechAllGroup)); 

	return returnValue;
}

void mechAllGroup_print(FILE *filePtr, mechAllGroup *aMechAllGroup)
{
	int i;

	fprintf(filePtr, "  Total no. of groups : %d\n", aMechAllGroup->noOfGroup);
	for (i = 0; i < aMechAllGroup->noOfGroup; i++)
	{
		fprintf(filePtr, " No. %d \n", i);
		mechGroup_print(filePtr, aMechAllGroup->grouping[i]);
	}
}

void mechAllGroup_free(mechAllGroup *aMechAllGroup)
{
	int i;

	int noOfGroup;
	mechGroup **grouping;

	noOfGroup = aMechAllGroup->noOfGroup;
	grouping = aMechAllGroup->grouping;

	for (i = 0; i < noOfGroup; i++)
	{
		mechGroup_free(grouping[i]);
	}
	if (noOfGroup > 0)
	{
		free(grouping);
	}

	free(aMechAllGroup);
}

void mechAllGroup_insert(mechAllGroup *aMechAllGroup, mechGroup **aMechGroup)
{
	mechGroup *oneMechGroup;

	int noOfGrouping;

	oneMechGroup = (*aMechGroup);

	noOfGrouping = aMechAllGroup->noOfGroup;

	if (noOfGrouping == 0)
	{
		aMechAllGroup->grouping = (mechGroup **) malloc(sizeof(mechGroup *));
	}
	else
	{
		aMechAllGroup->grouping = (mechGroup **) realloc(aMechAllGroup->grouping, sizeof(mechGroup *)*(noOfGrouping+1));
	}

	aMechAllGroup->grouping[noOfGrouping] = oneMechGroup;
	aMechAllGroup->noOfGroup++;
}

void mechAllGroup_remove(mechAllGroup *aMechAllGroup, int index)
{
	mechGroup **grouping;

	int noOfGroup;

	mechGroup **newGrouping;

	noOfGroup = aMechAllGroup->noOfGroup;
	grouping = aMechAllGroup->grouping;

	if (noOfGroup == 1)
	{
		newGrouping = NULL;
	}
	else
	{
		newGrouping = (mechGroup **) malloc(sizeof(mechGroup *)*(noOfGroup-1));

		if (index > 0)
		{
			memcpy(newGrouping, grouping, sizeof(mechGroup *)*index);
		}
		if (noOfGroup-1-index > 0)
		{
			memcpy(&(newGrouping[index]), &(grouping[index+1]), sizeof(mechGroup *)*(noOfGroup-1-index));
		}

	}

	free(grouping);
	aMechAllGroup->noOfGroup--;
	aMechAllGroup->grouping = newGrouping;
}

// -1: cannot find the coverage for this dimension
int mech_findCoverage_dim(dimension *aDim, generalizedDimTable *genDimTable)
{
	int returnValue;

	int noOfLevel;
	generalizedLevelTable *gen;
	generalizedLevelTable *genLevelTable;

	int level, value;
	int noOfPoss;

	int i;

	generalizedPair *genPairPtr;

	noOfLevel = genDimTable->noOfLevel;
	gen = genDimTable->gen;

	level = aDim->level;
	value = aDim->value;

	returnValue = -1;

	if (level != noOfLevel-1)
	{
		genLevelTable = &(gen[level]);
		noOfPoss = genLevelTable->noOfPoss;
		genPairPtr = genLevelTable->gen;
		for (i = 0; i < noOfPoss; i++)
		{
			if (genPairPtr[i].bottomValue == value)
			{
				returnValue = genPairPtr[i].coverage;
				break;
			}
		}
	}
	else
	{
		returnValue = genDimTable->coverage;
	}

	return returnValue;
}

// -1: cannot find
int mech_findParentCoverage_dim(generalizedDimTable *genDimTable, dimension *dim1, dimension *dim2)
{
	int returnValue;

	int noOfLevel;

	dimension *tempDim1;
	dimension *tempDim2;

	int level;
	//int level, value;
	//int i;

	//int noOfPoss;

	generalizedLevelTable *genLevelTable;
	generalizedLevelTable *gen;

	//generalizedPair *genPairPtr;

	returnValue = -1;

	tempDim1 = (dimension *) malloc(sizeof(dimension));
	tempDim2 = (dimension *) malloc(sizeof(dimension));

	memcpy(tempDim1, dim1, sizeof(dimension));
	memcpy(tempDim2, dim2, sizeof(dimension));

	noOfLevel = genDimTable->noOfLevel;
	gen = genDimTable->gen;

	// tempDim1 and tempDim2 should have the same level = 0 (ground level)
	while (tempDim1->value != tempDim2->value)
	{
		// generalize tempDim1
		level = tempDim1->level;
		genLevelTable = &(gen[level]);
		generalizeDimension(tempDim1, genLevelTable);

		// generalize tempDim2
		level = tempDim2->level;
		genLevelTable = &(gen[level]);
		generalizeDimension(tempDim2, genLevelTable);
	}

	// find the coverage
	returnValue = mech_findCoverage_dim(tempDim1, genDimTable);
	/*
	level = tempDim1->level;
	value = tempDim1->value;
	if (level != noOfLevel-1)
	{
		genLevelTable = &(gen[level]);
		noOfPoss = genLevelTable->noOfPoss;
		genPairPtr = genLevelTable->gen;
		for (i = 0; i < noOfPoss; i++)
		{
			if (genPairPtr[i].bottomValue == value)
			{
				returnValue = genPairPtr[i].coverage;
				break;
			}
		}
	}
	else
	{
		returnValue = genDimTable->coverage;
	}*/

	free(tempDim1);
	free(tempDim2);

	return returnValue;
}

//int mech_findParentCoverage_dim(generalizedDimTable *genDimTable, dimension *dim1, dimension *dim2)
double mech_calculateDistanceBetweenTwoClass(int id1, int id2, data *dataSet, int noOfQID, generalizedAllTable *genAllTable)
{
	double returnValue;
	double weightA;

	generalizedDimTable *genDimTable;

	tuple *tuple1;
	tuple *tuple2;

	dimension *content1;
	dimension *content2;

	dimension *dim1;
	dimension *dim2;


	int i;
	int parentCoverage;

	// assume uniform weighing
	weightA = 1.0/noOfQID;

	tuple1 = &(dataSet->tupleList[id1]);
	tuple2 = &(dataSet->tupleList[id2]);

	content1 = tuple1->content;
	content2 = tuple2->content;


	returnValue = 0;
	for (i = 0; i < noOfQID; i++)
	{
		dim1 = &(content1[i]);
		dim2 = &(content2[i]);
		genDimTable = &(genAllTable->genDimTable[i]);

		parentCoverage = mech_findParentCoverage_dim(genDimTable, dim1, dim2);
		returnValue += weightA*(parentCoverage-1.0)/(genDimTable->coverage-1.0);
	}

	return returnValue;
}

double mech_calculateDistanceBetweenClassMechGroup(int id, mechGroup *aMechGroup, data *dataSet, int noOfQID, generalizedAllTable *genAllTable)
{
	double returnValue;
	double temp;

	int noOfMember;
	int *member;
	int i;

	int id1, id2;

	noOfMember = aMechGroup->noOfMember;
	member = aMechGroup->member;

	id1 = id;

	returnValue = 0;
	for (i = 0; i < noOfMember; i++)
	{
		id2 = member[i];
		
		temp = mech_calculateDistanceBetweenTwoClass(id1, id2, dataSet, noOfQID, genAllTable);
		returnValue += temp;
	}

	returnValue = returnValue/noOfMember;

	return returnValue;
}


// assumme dim1 and dim2 in the ground value
int mech_isSameQID(dimension *dim1, dimension *dim2, int noOfQID)
{
	int returnValue;

	int i;

	returnValue = 1;
	for (i = 0; i < noOfQID; i++)
	{
		if (dim1[i].value != dim2[i].value)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

mechGroup *mech_findNeighborParticipant(mechGroup *aMechGroup, equivalenceClass *aEqClass, data *dataSet, int noOfQID, generalizedAllTable *genAllTable, int noOfValuInSenAttr)
{
	mechGroup *returnValue;
	int i, j;
	int isEqual;

	int noOfEqClass;
	equivalenceTuple **eqTuple;
	equivalenceTuple *aEqTuple;

	int noOfMember;
	int *member;

	dimension *dim1;
	dimension *dim2;

	int *isChosen;

	double minDist;
	double *dist;

	double MAX_DIST = 9999999;

	int id1, id2;

	int dummy;

	returnValue = mechGroup_new(noOfValuInSenAttr);

	noOfEqClass = aEqClass->noOfEqClass;
	eqTuple = aEqClass->eqClass;

	noOfMember = aMechGroup->noOfMember;
	member = aMechGroup->member;

	isChosen = (int *) malloc(sizeof(int)*noOfEqClass);
	memset(isChosen, 0, sizeof(int)*noOfEqClass);

	dist = (double *) malloc(sizeof(double)*noOfEqClass);
	memset(dist, 0, sizeof(double)*noOfEqClass);

	minDist = MAX_DIST;

	for (i = 0; i < noOfEqClass; i++)
	{
//if (i == 305) 
//{
//	printf("%d\n", i);
//}
		aEqTuple = eqTuple[i];
		// has non-sen tuples?
		if (aEqTuple->noOfMember-aEqTuple->noOfSen > 0)
		{
			// check whether particpant

			dim1 = aEqTuple->content;
			for (j = 0; j < noOfMember; j++)
			{
				id2 = member[j];
				dim2 = dataSet->tupleList[id2].content;
				
				isEqual = mech_isSameQID(dim1, dim2, noOfQID);

				if (isEqual == 1)
				{
					// parcipant
					isChosen[i] = 1;
					break;
				}
			}

			// not in mechGroup
			id1 = aEqTuple->member[0];
			if (isChosen[i] == 0)
			{
				// neighbor
				// check whether the distance is min
				dist[i] = mech_calculateDistanceBetweenClassMechGroup(id1, aMechGroup, dataSet, noOfQID, genAllTable);
				if (dist[i] < minDist)
				{
					minDist = dist[i];
				}
			}
		}
	}

	// neighbor
	for (i = 0; i < noOfEqClass; i++)
	{
		aEqTuple = eqTuple[i];

		if (aEqTuple->noOfMember-aEqTuple->noOfSen > 0)
		{
			if (minDist == dist[i])
			{
				isChosen[i] = 1;
			}
		}
	}

	// returnValue
	for (i = 0; i < noOfEqClass; i++)
	{
		if (isChosen[i] == 1)
		{
			dummy = 0;
			mech_insert(returnValue, i, 0, dummy);
		}
	}

	free(isChosen);
	free(dist);

	return returnValue;
}

// return the id of the tuple removed
int removeMemberFromEqTuple(equivalenceTuple *eqTuple, int index)
{
	int returnValue;
	int *newMember;
	int *newIsSen;

	int noOfMember;

	noOfMember = eqTuple->noOfMember;
	if (noOfMember == 1)
	{
		eqTuple->noOfMember--;
		if (eqTuple->isSen[index] == 1)
		{
			eqTuple->noOfSen--;
		}

		returnValue = eqTuple->member[index];

		newMember = NULL;
		newIsSen = NULL;

		free(eqTuple->member);
		free(eqTuple->isSen);

		eqTuple->member = newMember;
		eqTuple->isSen = newIsSen;
	}
	else
	{
		eqTuple->noOfMember--;
		if (eqTuple->isSen[index] == 1)
		{
			eqTuple->noOfSen--;
		}

		returnValue = eqTuple->member[index];

		newMember = (int *) malloc(sizeof(int)*(noOfMember-1));
		newIsSen = (int *) malloc(sizeof(int)*(noOfMember-1));

		if (index > 0)
		{
			memcpy(newMember, eqTuple->member, sizeof(int)*index);
			memcpy(newIsSen, eqTuple->isSen, sizeof(int)*index);
		}

		if (noOfMember-index-1 > 0)
		{
			memcpy(&(newMember[index]), &(eqTuple->member[index+1]), sizeof(int)*(noOfMember-index-1));
			memcpy(&(newIsSen[index]), &(eqTuple->isSen[index+1]), sizeof(int)*(noOfMember-index-1));
		}

		free(eqTuple->member);
		free(eqTuple->isSen);

		eqTuple->member = newMember;
		eqTuple->isSen = newIsSen;

		
	}

	return returnValue;

	
}

// return the ID removed
int mech_removeNonSenIDFromEqTuple(equivalenceTuple *eqTuple)
{
	int returnValue;
	int i;

	for (i = 0; i < eqTuple->noOfMember; i++)
	{
		if (eqTuple->isSen[i] == 0)
		{
			// remove 
			returnValue = removeMemberFromEqTuple(eqTuple, i);
			
			break;
		}
	}

	return returnValue;
}

// return the ID removed
int mech_removeSenIDFromEqTuple(equivalenceTuple *eqTuple)
{
	int returnValue;
	int i;

	for (i = 0; i < eqTuple->noOfMember; i++)
	{
		if (eqTuple->isSen[i] != 0)
		{
			// remove 
			returnValue = removeMemberFromEqTuple(eqTuple, i);
			
			break;
		}
	}

	return returnValue;
}

mechGroup *transformFromEqTupleToMechGroup(equivalenceTuple *eqTuple, int noOfValuInSenAttr)
{
	mechGroup *returnValue;

	int i;
	int id, isSen;

	int senValue;

	returnValue = mechGroup_new(noOfValuInSenAttr);

	for (i = 0; i < eqTuple->noOfMember; i++)
	{
		id = eqTuple->member[i];
		isSen = eqTuple->isSen[i];
		// assume ground value
		senValue = eqTuple->newSenValue[i].value;

		mech_insert(returnValue, id, isSen, senValue);
	}

	return returnValue;
}

mechAllGroup *transformFromEqClassToMechAllGroup(equivalenceClass *eqClass, int noOfValueInSenAttr)
{
	mechAllGroup *returnValue;
	int noOfEqTuple;
	equivalenceTuple **eqList;

	mechGroup *aMechGroup;

	int i;

	returnValue = mechAllGroup_new();

	noOfEqTuple = eqClass->noOfEqClass;
	eqList = eqClass->eqClass;

	for (i = 0; i < noOfEqTuple; i++)
	{
		aMechGroup = transformFromEqTupleToMechGroup(eqList[i], noOfValueInSenAttr);

		mechAllGroup_insert(returnValue, &aMechGroup);
	}
	return returnValue;
}

void mech_mergeAllMechGroupWithSameQID(mechAllGroup *aMechAllGroup, data *dataSet, int noOfQID)
{
	int i, j, k;

	int dummy;

	int noOfGroup;

	int isEqual;
	int *isAllEqual;
	mechGroup **grouping;

	int id, isSen;

	noOfGroup = aMechAllGroup->noOfGroup;
	grouping = aMechAllGroup->grouping;

	isAllEqual = (int *) malloc(sizeof(int)*noOfGroup);
	memset(isAllEqual, 0, sizeof(int)*noOfGroup);

	for (i = 0; i < noOfGroup; i++)
	{
		isAllEqual[i] = mech_isAllSameQIDInAMechGroup(grouping[i], dataSet, noOfQID);
	}

	for (i = noOfGroup-1; i >= 0; i--)
	{
		if (isAllEqual[i] == 1)
		{
			for (j = i-1; j >= 0; j--)
			{
				if (isAllEqual[j] == 1)
				{
if ((i == 11) && (j == 7))
{
	printf("");
}
					grouping = aMechAllGroup->grouping;
					isEqual = mech_isAllSameQIDTwoMechGroup(grouping[i], grouping[j], dataSet, noOfQID);
					if (isEqual == 1)
					{
						
						// insert all members in mechGroup i to mechGroup j

						for (k = 0; k < grouping[i]->noOfMember; k++)
						{
							id = grouping[i]->member[k];
							isSen = grouping[i]->isSen[k];
							dummy = 0;
							mech_insert(grouping[j], id, isSen, dummy);
						}

						// remove mechGroup i
						mechGroup_free(grouping[i]);
						mechAllGroup_remove(aMechAllGroup, i);						
						break;
					}
				}
			}
		}
	}

	free(isAllEqual);
}

twoTable *twoTable_new(data *dataSet, int noOfGroup)
{
	twoTable *returnValue;
	data *tempData;

	int i;

	returnValue = (twoTable *) malloc(sizeof(twoTable));
	
	// qid
	tempData = (data *) malloc(sizeof(data));
	tempData->noOfDim = dataSet->noOfDim;
	tempData->noOfTuple = dataSet->noOfTuple;
	tempData->tupleList = (tuple *) malloc(sizeof(tuple)*tempData->noOfTuple);
	tempData->dimSize = (int *) malloc(sizeof(int)*tempData->noOfDim);

	for (i = 0; i < tempData->noOfTuple; i++)
	{
		tempData->tupleList[i].content = (dimension *) malloc(sizeof(dimension)*tempData->noOfDim);
	}

	memcpy(tempData->dimSize, dataSet->dimSize, sizeof(int)*tempData->noOfDim);
	tempData->dimSize[dataSet->noOfDim-1] = noOfGroup;
	returnValue->qidTable = tempData;

	// sen
	tempData = (data *) malloc(sizeof(data));
	tempData->noOfDim = 2;
	tempData->noOfTuple = dataSet->noOfTuple;
	tempData->tupleList = (tuple *) malloc(sizeof(tuple)*tempData->noOfTuple);
	tempData->dimSize = (int *) malloc(sizeof(int)*tempData->noOfDim);

	for (i = 0; i < tempData->noOfTuple; i++)
	{
		tempData->tupleList[i].content = (dimension *) malloc(sizeof(dimension)*tempData->noOfDim);
	}

	tempData->dimSize[0] = noOfGroup;
	// assume the last dim is sensitive attr.
	tempData->dimSize[1] = dataSet->dimSize[dataSet->noOfDim-1];
	returnValue->senTable = tempData;

	return returnValue;
}


void insertToQIDTable(data *qidTable, int index, int gid, dimension *content, int noOfOrgDim)
{
	tuple *aTuple;

	aTuple = &(qidTable->tupleList[index]);
	memcpy(aTuple->content, content, sizeof(dimension)*(noOfOrgDim-1));
	aTuple->content[noOfOrgDim-1].value = gid;
	aTuple->content[noOfOrgDim-1].level = 0;
}

// assume the last dim is sen. attr.
void insertToSenTable(data *qidTable, int index, int gid, dimension *content, int noOfOrgDim)
{
	tuple *aTuple;

	aTuple = &(qidTable->tupleList[index]);
	aTuple->content[0].value = gid;
	aTuple->content[0].level = 0;
	aTuple->content[1].value = content[noOfOrgDim-1].value;
	aTuple->content[1].level = content[noOfOrgDim-1].level;
}

twoTable *generateTwoTableFromMechGroup(mechAllGroup *aMechAllGroup, data *dataSet)
{
	twoTable *returnValue;

	int noOfGroup;
	mechGroup **grouping;

	int noOfMember;
	int *member;

	int i, j;

	int id;

	tuple *aTuple;

	data *qidTable;
	data *senTable;

	int count;
	int gid;
	int noOfOrgDim;

	noOfGroup = aMechAllGroup->noOfGroup;
	grouping = aMechAllGroup->grouping;

	returnValue = twoTable_new(dataSet, noOfGroup);

	qidTable = returnValue->qidTable;
	senTable = returnValue->senTable;

	noOfOrgDim = dataSet->noOfDim;

	count = 0;
	for (i = 0; i < noOfGroup; i++)
	{
		noOfMember = grouping[i]->noOfMember;
		member = grouping[i]->member;
		for (j = 0; j < noOfMember; j++)
		{
			id = member[j];

			aTuple = &(dataSet->tupleList[id]);

			gid = i;

			insertToQIDTable(qidTable, count, gid, aTuple->content, noOfOrgDim);

			insertToSenTable(senTable, count, gid, aTuple->content, noOfOrgDim);

			count++;
		}
	}

	return returnValue;
}

int isMech(equivalenceClass *proEqClass, configure *conf, int noOfValueInSenAttr)
{
	int returnValue;
	int i, j;
	int *totalNoOfSen;

	int senStartNo, senEndNo;

	senStartNo = conf->senStartNo;
	senEndNo = conf->senEndNo;

	totalNoOfSen = (int *) malloc(sizeof(int)*noOfValueInSenAttr);
	memset(totalNoOfSen, 0, sizeof(int)*noOfValueInSenAttr);

	for (i = 0; i < proEqClass->noOfEqClass; i++)
	{
		for (j = senStartNo; j <= senEndNo ; j++)
		{
			totalNoOfSen[j] += proEqClass->eqClass[i]->noOfEachSen[j];
		}
	}

	returnValue = 1;
	for (j = senStartNo; j <= senEndNo; j++)
	{
		if (totalNoOfSen[j]*conf->d <= conf->noOfTuple)
		{
			// do nothing
		}
		else
		{
			returnValue =  0;
			break;
		}
	}
	free(totalNoOfSen);

	return returnValue;
}

void printData(FILE *filePtr, data *dataSet)
{
	int i, j;

	for (i = 0; i < dataSet->noOfTuple; i++)
	{
		fprintf(filePtr, " %d. (", i);
		for (j = 0; j < dataSet->noOfDim; j++)
		{
			fprintf(filePtr, "%d", dataSet->tupleList[i].content[j]);
			if (j != dataSet->noOfDim-1)
			{
				fprintf(filePtr, ", ");
			}
			
		}
		fprintf(filePtr, ")\n");
	}
}

void printTwoTable(FILE *filePtr, twoTable *aTwoTable)
{
	data *qidTable;
	data *senTable;
	
	qidTable = aTwoTable->qidTable;
	senTable = aTwoTable->senTable;

	fprintf(filePtr, "QID Table\n");
	fprintf(filePtr, "=========\n");
	printData(filePtr, qidTable);

	fprintf(filePtr, "sen Table\n");
	fprintf(filePtr, "=========\n");
	printData(filePtr, senTable);
}

// noOfDim is used for restriction for the consideration of attr.
// assume compare the ground values
int isMatchCondition(dimension *aDimension, condition *aCondition, int noOfDim)
{
	int returnValue;
	int i, j;

	int *dimNo;
	predicate **predicateList;

	int noOfValue;
	int *value;

	int aValue;

	int *isMatch;

	int noOfPredicate;

	int aDimNo;

	

	

	dimNo = aCondition->dimNo;
	predicateList = aCondition->predicateList;
	noOfPredicate = aCondition->noOfPredicate;

	isMatch = (int *) malloc(sizeof(int)*noOfPredicate);
	memset(isMatch, 0, sizeof(int)*noOfPredicate);

	for (i = 0; i < noOfPredicate; i++)
	{
		aDimNo = dimNo[i];
		if (aDimNo < noOfDim)
		{
			noOfValue = predicateList[i]->noOfValue;
			value = predicateList[i]->value;

			isMatch[i] = 0;
			for (j = 0; j < noOfValue; j++)
			{
				aValue = value[j]; 

				if (aValue == aDimension[aDimNo].value)
				{
					isMatch[i] = 1;
					break;
				}
			}
		}
	}

	returnValue = 1;
	for (i = 0; i < noOfPredicate; i++)
	{
		aDimNo = dimNo[i];
		if (aDimNo < noOfDim)
		{
			if (isMatch[i] == 0)
			{
				returnValue = 0;
				break;
			}
		}
	}

	free(isMatch);

	return returnValue;
}

// assume only sen. and non-sen
double calculateCountMechGroup(mechAllGroup *aMechAllGroup, condition *aCondition, data *dataSet, int noOfQID)
{
	int noOfGroup;
	mechGroup **grouping;
	mechGroup *aMechGroup;

	int i, j;
	int id;

	dimension *aDimension;

	int isMatch;

	int totalResCount;
	int dimSize;

	double count, countForAGroup;

	predicate *aPredicate;

	int *temp;
	int noOfDistinctValue;

	noOfGroup = aMechAllGroup->noOfGroup;
	grouping = aMechAllGroup->grouping;

	count = 0;
	for (i = 0; i < noOfGroup; i++)
	{
		aMechGroup = grouping[i];
		countForAGroup = 0;
		for (j = 0; j < aMechGroup->noOfMember; j++)
		{
			id = aMechGroup->member[j];

			aDimension = dataSet->tupleList[id].content;

			// use noOfQID to filter the sen. attr cal.
			isMatch = isMatchCondition(aDimension, aCondition, noOfQID);

			if (isMatch == 1)
			{
				countForAGroup++;
			}
		}

		// special case for sen.
		aPredicate = aCondition->predicateList[aCondition->noOfPredicate-1];
		dimSize = dataSet->dimSize[noOfQID];
		temp = (int *) malloc(sizeof(int)*dimSize);
		memset(temp, 0, sizeof(int)*dimSize);
		for (j = 0; j < aPredicate->noOfValue; j++)
		{
			temp[aPredicate->value[j]] = 1;
		}
		noOfDistinctValue  = 0;
		for (j = 0; j < dimSize; j++)
		{
			noOfDistinctValue += temp[j];
		}

		//
		totalResCount = 0;
		for (j = 0; j < dimSize; j++)
		{
			if (temp[j] == 1)
			{
				totalResCount += aMechGroup->noOfEachSen[j];
			}
		}
		countForAGroup *= (totalResCount/((double)aMechGroup->noOfMember));
		
		/*
		if (noOfDistinctValue == 1)
		{
			if (aPredicate->value[0] == 0)
			{
				countForAGroup *= ((aMechGroup->noOfMember-aMechGroup->noOfSen)/((double)aMechGroup->noOfMember));
			}
			else
			{
				countForAGroup *= ((aMechGroup->noOfSen)/((double)aMechGroup->noOfMember));
			}
		}
		else if (noIfDistinctValue == 2)
		{
			// there are two values
			countForAGroup *= 1;
		}
		else
		{
			coutnForAGroup *= 0;
		}
		*/
		free(temp);

		count += countForAGroup;
	}

	return count;
}

double calculateCountDataSet(condition *aCondition, data *dataSet)
{
	int noOfTuple;
	tuple *tupleList;

	int isMatch;

	int i;
	int noOfDim;

	int count;

	dimension *aDimension;

	noOfDim = dataSet->noOfDim;

	noOfTuple = dataSet->noOfTuple;
	tupleList = dataSet->tupleList;

	count = 0;

	for (i = 0; i < noOfTuple; i++)
	{
		aDimension = tupleList[i].content;

		isMatch = isMatchCondition(aDimension, aCondition, noOfDim);

		if (isMatch == 1)
		{
			count++;
		}
	}

	return count;
}

double calculateAvgRelativeError(mechAllGroup *aMechAllGroup, allCondition *aAllCondition, data *dataSet, int noOfQID)
{
	double returnValue;
	double est, act;

	condition *aCondition;
	double tempValue;
	int i;

	returnValue = 0;
	for (i = 0; i < aAllCondition->noOfCondition; i++)
	{
		aCondition = aAllCondition->conditionList[i];

		est = calculateCountMechGroup(aMechAllGroup, aCondition, dataSet, noOfQID);
		act = calculateCountDataSet(aCondition, dataSet);

		if (act != 0)
		{
			if (act > est)
			{
				tempValue = (act-est)/act;
			}
			else
			{
				tempValue = (est-act)/act;
			}
			returnValue += tempValue;
		}
	}
	returnValue = returnValue/aAllCondition->noOfCondition;

	return returnValue;
}

void mechMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable)
{
	FILE *debugFilePtr;

	int isAlphaK;

	int kAnonymity;
	int noOfQID;

	summary *synopsis;
	equivalenceClass *proEqClass;
	//equivalenceClass *finalEqClass;
	mechAllGroup *finalMechAllGroup;

	int noOfValueInSenAttr;

	//int noOfGreatestPossRemovalCount;

	//int *listOfPotentialTrunk;
	//int noOfPotentialTrunk;

	//int prevNoOfFinalEqClass;

	//int noOfRemain;

	//int bestAttrNo;

	//int iterationNo;

	//double distortion;

	//double *dist;

	//Hash *hash;
	//DistStorage *distStorage;
	int noOfSlot;
	float delta;

	float maxValue;

	int dataID;
	int processEqID;
	mechGroup *aMechGroup;
	int isSen;

	double avgRelError;

	int i;
	int noOfProEqClass;

	int randIndex;

	mechGroup *L;

	int chosenDataID;
	int chosenEqID;

	allCondition *aAllCondition;

	twoTable *finalTable;
	FILE *filePtr;

	int iterationNo;

	int dummy;

	char *avgRelErrorFilename = "error.txt";
	char *timeFilename = "time.txt";

	int queryNoOfDim = conf->queryNoOfDim;
	double s = conf->s;
	int noOfCondition = conf->noOfCondition;

#ifndef _WIN32
	float  userTime, sysTime;
	struct rusage myTime1, myTime2;

	getrusage(RUSAGE_SELF,&myTime1);
#endif

	noOfQID = conf->noOfDim-conf->noOfSenAttr;
	kAnonymity = conf->K;

	if (conf->noOfTuple > 200)
	{
		noOfSlot = ((conf->noOfTuple*conf->noOfTuple)/200)-1;
	}
	else
	{
		noOfSlot = 5;
	}
	delta = 5.0f;

	maxValue = (float) (conf->noOfDim*2)*conf->noOfTuple/conf->K;

	debugFilePtr = (FILE *) fopen("debug.txt", "w");
	g_debugFilePtr = debugFilePtr;

	printf("Random Algorithm....\n");

	// preprocessing
	//printf("Preprocessing....\n");
	//prepProcessDataSet(dataSet, conf->K, genAllTable);

	// create summary
	printf("Creating Summary....\n");
	synopsis = createSummary(dataSet);

	// printing
	//fprintf(debugFilePtr, "############## Summary ##################\n");
	//printf("############## Summary ##################\n");
	//printSummary(debugFilePtr, synopsis, conf->noOfDim);

	// check whether it is alphaK
	//isAlphaK = isAlphaKAnonymity(dataSet, synopsis, conf);
	//isAlphaK = 1;

	// generate equivalence class from the data
	printf("Generating eq. class from Data....\n");
	proEqClass = generateEquivalenceClass(dataSet, synopsis, conf); 
	noOfValueInSenAttr = dataSet->dimSize[noOfQID];
	isAlphaK = isMech(proEqClass, conf, noOfValueInSenAttr);

	// 
	if (isAlphaK == 0)
	{
		printf("This data set cannot satisfy the alpha K anonymity!\n");
		exit(0);
	}
	else
	{
		// continue

		// generalize to the top level
		//proEqClass = top_generalize2TopLevel(dataSet, genAllTable);

		// generate a final mechAllGroup
		finalMechAllGroup = mechAllGroup_new();

		//fprintf(debugFilePtr, "############## proEqClass ##################\n");
		//printf("############## proEqClass ##################\n");
		//printEquivalenceClass(debugFilePtr, proEqClass, conf->noOfDim);

	

		noOfProEqClass = proEqClass->noOfEqClass;
		processEqID = -1;
		for (i = 0; i < noOfProEqClass; i++)
		{
			if (proEqClass->eqClass[i]->noOfSen > 0)
			{
				processEqID = i;
				break;
			}
		}


		iterationNo = 0;

		while (processEqID != -1)
		{
			if (iterationNo%10 == 0)
			{
				printf("Iteration %d....\n", iterationNo);
			}
if (iterationNo == 626)
{
	printf("");
}

			// printing
			//fprintf(debugFilePtr, "Printing finalMechAllGroup ...\n");
			//fprintf(debugFilePtr, "-------------------------------\n");
			//mechAllGroup_print(debugFilePtr, finalMechAllGroup);

			//fprintf(debugFilePtr, "############## proEqClass ##################\n");
			//printf("############## proEqClass ##################\n");
			//printEquivalenceClass(debugFilePtr, proEqClass, conf->noOfDim);

			// remove sen. tuple from D
			dataID = mech_removeSenIDFromEqTuple(proEqClass->eqClass[processEqID]);
			isSen = 1;
			
			aMechGroup = mechGroup_new(noOfValueInSenAttr);

			// insert sen. tuple to E
			dummy = 0;
			mech_insert(aMechGroup, dataID, isSen, dummy);
//while (1){
			//
			while (aMechGroup->noOfSen*kAnonymity > aMechGroup->noOfMember)
			{
				// find neighbor/participant
				L = mech_findNeighborParticipant(aMechGroup, proEqClass, dataSet, noOfQID, genAllTable, noOfValueInSenAttr);

				// randomly chosee
				randIndex = rand()%(L->noOfMember);

				chosenEqID = L->member[randIndex];

				chosenDataID = mech_removeNonSenIDFromEqTuple(proEqClass->eqClass[chosenEqID]);
				isSen = 0;

				// insert 
				dummy = 0;
				mech_insert(aMechGroup, chosenDataID, isSen, dummy);

				// free L
				mechGroup_free(L);
			}
//}
			// insert to finalMechAllGroup
			mechAllGroup_insert(finalMechAllGroup, &aMechGroup);

			// 
			noOfProEqClass = proEqClass->noOfEqClass;
			processEqID = -1;
			for (i = 0; i < noOfProEqClass; i++)
			{
				if (proEqClass->eqClass[i]->noOfSen > 0)
				{
					processEqID = i;
					break;
				}
			}

			iterationNo++;
		}


		// D contains all non-sensitive tuples
		for (i = 0; i < noOfProEqClass; i++)
		{
			if (proEqClass->eqClass[i]->noOfMember != 0)
			{
				// insert
				aMechGroup = transformFromEqTupleToMechGroup(proEqClass->eqClass[i], noOfValueInSenAttr);
				mechAllGroup_insert(finalMechAllGroup, &aMechGroup);
			}
		}

		// merge all groups with the same QID
		printf("Merging groups with the same QID...\n");
		mech_mergeAllMechGroupWithSameQID(finalMechAllGroup, dataSet, noOfQID);

		// printing
		//fprintf(debugFilePtr, "Printing finalMechAllGroup ...\n");
		//fprintf(debugFilePtr, "-------------------------------\n");
		//mechAllGroup_print(debugFilePtr, finalMechAllGroup);


		

	}

#ifndef _WIN32
	getrusage(RUSAGE_SELF, &myTime2);	
#endif
	//
	//allCondition *predicate_generateAllCondition(int noOfCondition, int queryNoOfDim, int noOfQID, double s, int *dimSize)
	printf("Generating predicates...\n");
	aAllCondition = predicate_generateAllCondition(noOfCondition, queryNoOfDim, noOfQID, s, dataSet->dimSize);

	
	// generate two tables
	printf("Generating two tables...\n");
	finalTable = generateTwoTableFromMechGroup(finalMechAllGroup, dataSet);

	// printing
	//printTwoTable(debugFilePtr, finalTable);

	// calculate the estimated count and actual count
	avgRelError = calculateAvgRelativeError(finalMechAllGroup, aAllCondition, dataSet, noOfQID);

	printf("Average relative error : %.2f\n", avgRelError);

	filePtr = (FILE *) fopen(avgRelErrorFilename, "w");
	fprintf(filePtr, "%.4f\n", avgRelError);
	fclose(filePtr);


#ifndef _WIN32
	userTime =
     		((float) (myTime2.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime2.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime2.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime2.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;

	printf("User time : %f seconds\n",userTime);
	printf("System time : %f seconds\n\n",sysTime);

	fprintf(debugFilePtr, "Execution Time\n");
	fprintf(debugFilePtr, "--------------\n");
	fprintf(debugFilePtr, "User time : %f seconds\n",userTime);
    fprintf(debugFilePtr, "System time : %f seconds\n\n",sysTime);


	filePtr = (FILE *) fopen(timeFilename, "w");
	fprintf(filePtr, "%.6f\n",userTime);
    fprintf(filePtr, "%.6f\n\n",sysTime);
	fclose(filePtr);


#endif

	fclose(debugFilePtr);
}



